/* Copyright 2022 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================*/

#ifndef MLIR_HLO_DIALECT_MHLO_TRANSFORMS_BUFFERIZABLE_OP_INTERFACE_IMPL_H
#define MLIR_HLO_DIALECT_MHLO_TRANSFORMS_BUFFERIZABLE_OP_INTERFACE_IMPL_H

#include <functional>
#include <memory>

#include "mlir/Dialect/Bufferization/IR/BufferizableOpInterface.h"

namespace mlir {
namespace mhlo {

/// Register the external models for bufferizing mhlo ops.
void registerBufferizableOpInterfaceExternalModels(DialectRegistry &registry);

}  // namespace mhlo
}  // namespace mlir

#endif  // MLIR_HLO_DIALECT_MHLO_TRANSFORMS_BUFFERIZABLE_OP_INTERFACE_IMPL_H
