# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorflow/core/protobuf/status.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorflow/core/protobuf/status.proto',
  package='tensorflow',
  syntax='proto3',
  serialized_options=_b('ZUgithub.com/tensorflow/tensorflow/tensorflow/go/core/protobuf/for_core_protos_go_proto'),
  serialized_pb=_b('\n%tensorflow/core/protobuf/status.proto\x12\ntensorflow\"\x0f\n\rDerivedStatus\"\xa0\x01\n\x11StackTracePayload\x12>\n\x0cstack_frames\x18\x01 \x03(\x0b\x32(.tensorflow.StackTracePayload.StackFrame\x1aK\n\nStackFrame\x12\x11\n\tfile_name\x18\x01 \x01(\t\x12\x13\n\x0bline_number\x18\x02 \x01(\x05\x12\x15\n\rfunction_name\x18\x03 \x01(\tBWZUgithub.com/tensorflow/tensorflow/tensorflow/go/core/protobuf/for_core_protos_go_protob\x06proto3')
)




_DERIVEDSTATUS = _descriptor.Descriptor(
  name='DerivedStatus',
  full_name='tensorflow.DerivedStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=68,
)


_STACKTRACEPAYLOAD_STACKFRAME = _descriptor.Descriptor(
  name='StackFrame',
  full_name='tensorflow.StackTracePayload.StackFrame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='file_name', full_name='tensorflow.StackTracePayload.StackFrame.file_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='line_number', full_name='tensorflow.StackTracePayload.StackFrame.line_number', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='function_name', full_name='tensorflow.StackTracePayload.StackFrame.function_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=156,
  serialized_end=231,
)

_STACKTRACEPAYLOAD = _descriptor.Descriptor(
  name='StackTracePayload',
  full_name='tensorflow.StackTracePayload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='stack_frames', full_name='tensorflow.StackTracePayload.stack_frames', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_STACKTRACEPAYLOAD_STACKFRAME, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=71,
  serialized_end=231,
)

_STACKTRACEPAYLOAD_STACKFRAME.containing_type = _STACKTRACEPAYLOAD
_STACKTRACEPAYLOAD.fields_by_name['stack_frames'].message_type = _STACKTRACEPAYLOAD_STACKFRAME
DESCRIPTOR.message_types_by_name['DerivedStatus'] = _DERIVEDSTATUS
DESCRIPTOR.message_types_by_name['StackTracePayload'] = _STACKTRACEPAYLOAD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DerivedStatus = _reflection.GeneratedProtocolMessageType('DerivedStatus', (_message.Message,), {
  'DESCRIPTOR' : _DERIVEDSTATUS,
  '__module__' : 'tensorflow.core.protobuf.status_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.DerivedStatus)
  })
_sym_db.RegisterMessage(DerivedStatus)

StackTracePayload = _reflection.GeneratedProtocolMessageType('StackTracePayload', (_message.Message,), {

  'StackFrame' : _reflection.GeneratedProtocolMessageType('StackFrame', (_message.Message,), {
    'DESCRIPTOR' : _STACKTRACEPAYLOAD_STACKFRAME,
    '__module__' : 'tensorflow.core.protobuf.status_pb2'
    # @@protoc_insertion_point(class_scope:tensorflow.StackTracePayload.StackFrame)
    })
  ,
  'DESCRIPTOR' : _STACKTRACEPAYLOAD,
  '__module__' : 'tensorflow.core.protobuf.status_pb2'
  # @@protoc_insertion_point(class_scope:tensorflow.StackTracePayload)
  })
_sym_db.RegisterMessage(StackTracePayload)
_sym_db.RegisterMessage(StackTracePayload.StackFrame)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
