/* Autogenerated by mlir-tblgen; don't manually edit */
#ifdef GEN_PASS_CLASSES

//===----------------------------------------------------------------------===//
// DedupBoundInputBindingPass
//===----------------------------------------------------------------------===//

template <typename DerivedT>
class DedupBoundInputBindingPassBase : public ::mlir::OperationPass<mlir::func::FuncOp> {
public:
  using Base = DedupBoundInputBindingPassBase;

  DedupBoundInputBindingPassBase() : ::mlir::OperationPass<mlir::func::FuncOp>(::mlir::TypeID::get<DerivedT>()) {}
  DedupBoundInputBindingPassBase(const DedupBoundInputBindingPassBase &other) : ::mlir::OperationPass<mlir::func::FuncOp>(other) {}

  /// Returns the command-line argument attached to this pass.
  static constexpr ::llvm::StringLiteral getArgumentName() {
    return ::llvm::StringLiteral("tf-saved-model-dedup-bound-input-binding-pass");
  }
  ::llvm::StringRef getArgument() const override { return "tf-saved-model-dedup-bound-input-binding-pass"; }

  ::llvm::StringRef getDescription() const override { return "Remove duplicate 'tf_saved_model.bound_input' bindings."; }

  /// Returns the derived pass name.
  static constexpr ::llvm::StringLiteral getPassName() {
    return ::llvm::StringLiteral("DedupBoundInputBindingPass");
  }
  ::llvm::StringRef getName() const override { return "DedupBoundInputBindingPass"; }

  /// Support isa/dyn_cast functionality for the derived pass class.
  static bool classof(const ::mlir::Pass *pass) {
    return pass->getTypeID() == ::mlir::TypeID::get<DerivedT>();
  }

  /// A clone method to create a copy of this pass.
  std::unique_ptr<::mlir::Pass> clonePass() const override {
    return std::make_unique<DerivedT>(*static_cast<const DerivedT *>(this));
  }

  /// Return the dialect that must be loaded in the context before this pass.
  void getDependentDialects(::mlir::DialectRegistry &registry) const override {
    
  }

  /// Explicitly declare the TypeID for this class. We declare an explicit private
  /// instantiation because Pass classes should only be visible by the current
  /// library.
  MLIR_DEFINE_EXPLICIT_INTERNAL_INLINE_TYPE_ID(DedupBoundInputBindingPassBase<DerivedT>)

protected:
};

//===----------------------------------------------------------------------===//
// FreezeAssetsPass
//===----------------------------------------------------------------------===//

template <typename DerivedT>
class FreezeAssetsPassBase : public ::mlir::OperationPass<ModuleOp> {
public:
  using Base = FreezeAssetsPassBase;

  FreezeAssetsPassBase() : ::mlir::OperationPass<ModuleOp>(::mlir::TypeID::get<DerivedT>()) {}
  FreezeAssetsPassBase(const FreezeAssetsPassBase &other) : ::mlir::OperationPass<ModuleOp>(other) {}

  /// Returns the command-line argument attached to this pass.
  static constexpr ::llvm::StringLiteral getArgumentName() {
    return ::llvm::StringLiteral("tf-saved-model-freeze-assets");
  }
  ::llvm::StringRef getArgument() const override { return "tf-saved-model-freeze-assets"; }

  ::llvm::StringRef getDescription() const override { return "Freeze tf_saved_model.asset's in func bodies."; }

  /// Returns the derived pass name.
  static constexpr ::llvm::StringLiteral getPassName() {
    return ::llvm::StringLiteral("FreezeAssetsPass");
  }
  ::llvm::StringRef getName() const override { return "FreezeAssetsPass"; }

  /// Support isa/dyn_cast functionality for the derived pass class.
  static bool classof(const ::mlir::Pass *pass) {
    return pass->getTypeID() == ::mlir::TypeID::get<DerivedT>();
  }

  /// A clone method to create a copy of this pass.
  std::unique_ptr<::mlir::Pass> clonePass() const override {
    return std::make_unique<DerivedT>(*static_cast<const DerivedT *>(this));
  }

  /// Return the dialect that must be loaded in the context before this pass.
  void getDependentDialects(::mlir::DialectRegistry &registry) const override {
    
  }

  /// Explicitly declare the TypeID for this class. We declare an explicit private
  /// instantiation because Pass classes should only be visible by the current
  /// library.
  MLIR_DEFINE_EXPLICIT_INTERNAL_INLINE_TYPE_ID(FreezeAssetsPassBase<DerivedT>)

protected:
};

//===----------------------------------------------------------------------===//
// FreezeGlobalTensorsPass
//===----------------------------------------------------------------------===//

template <typename DerivedT>
class FreezeGlobalTensorsPassBase : public ::mlir::OperationPass<ModuleOp> {
public:
  using Base = FreezeGlobalTensorsPassBase;

  FreezeGlobalTensorsPassBase() : ::mlir::OperationPass<ModuleOp>(::mlir::TypeID::get<DerivedT>()) {}
  FreezeGlobalTensorsPassBase(const FreezeGlobalTensorsPassBase &other) : ::mlir::OperationPass<ModuleOp>(other) {}

  /// Returns the command-line argument attached to this pass.
  static constexpr ::llvm::StringLiteral getArgumentName() {
    return ::llvm::StringLiteral("tf-saved-model-freeze-global-tensors");
  }
  ::llvm::StringRef getArgument() const override { return "tf-saved-model-freeze-global-tensors"; }

  ::llvm::StringRef getDescription() const override { return "Freeze tf_saved_model.global_tensor's in func bodies."; }

  /// Returns the derived pass name.
  static constexpr ::llvm::StringLiteral getPassName() {
    return ::llvm::StringLiteral("FreezeGlobalTensorsPass");
  }
  ::llvm::StringRef getName() const override { return "FreezeGlobalTensorsPass"; }

  /// Support isa/dyn_cast functionality for the derived pass class.
  static bool classof(const ::mlir::Pass *pass) {
    return pass->getTypeID() == ::mlir::TypeID::get<DerivedT>();
  }

  /// A clone method to create a copy of this pass.
  std::unique_ptr<::mlir::Pass> clonePass() const override {
    return std::make_unique<DerivedT>(*static_cast<const DerivedT *>(this));
  }

  /// Return the dialect that must be loaded in the context before this pass.
  void getDependentDialects(::mlir::DialectRegistry &registry) const override {
    
  }

  /// Explicitly declare the TypeID for this class. We declare an explicit private
  /// instantiation because Pass classes should only be visible by the current
  /// library.
  MLIR_DEFINE_EXPLICIT_INTERNAL_INLINE_TYPE_ID(FreezeGlobalTensorsPassBase<DerivedT>)

protected:
  ::mlir::Pass::Option<bool> allow_mutable_tensors{*this, "allow-mutable-tensors", ::llvm::cl::desc("Allows mutable tensors to be in the graph."), ::llvm::cl::init(false)};
};

//===----------------------------------------------------------------------===//
// OptimizeGlobalTensorsPass
//===----------------------------------------------------------------------===//

template <typename DerivedT>
class OptimizeGlobalTensorsPassBase : public ::mlir::OperationPass<ModuleOp> {
public:
  using Base = OptimizeGlobalTensorsPassBase;

  OptimizeGlobalTensorsPassBase() : ::mlir::OperationPass<ModuleOp>(::mlir::TypeID::get<DerivedT>()) {}
  OptimizeGlobalTensorsPassBase(const OptimizeGlobalTensorsPassBase &other) : ::mlir::OperationPass<ModuleOp>(other) {}

  /// Returns the command-line argument attached to this pass.
  static constexpr ::llvm::StringLiteral getArgumentName() {
    return ::llvm::StringLiteral("tf-saved-model-optimize-global-tensors");
  }
  ::llvm::StringRef getArgument() const override { return "tf-saved-model-optimize-global-tensors"; }

  ::llvm::StringRef getDescription() const override { return "Optimize tf_saved_model.global_tensor's."; }

  /// Returns the derived pass name.
  static constexpr ::llvm::StringLiteral getPassName() {
    return ::llvm::StringLiteral("OptimizeGlobalTensorsPass");
  }
  ::llvm::StringRef getName() const override { return "OptimizeGlobalTensorsPass"; }

  /// Support isa/dyn_cast functionality for the derived pass class.
  static bool classof(const ::mlir::Pass *pass) {
    return pass->getTypeID() == ::mlir::TypeID::get<DerivedT>();
  }

  /// A clone method to create a copy of this pass.
  std::unique_ptr<::mlir::Pass> clonePass() const override {
    return std::make_unique<DerivedT>(*static_cast<const DerivedT *>(this));
  }

  /// Return the dialect that must be loaded in the context before this pass.
  void getDependentDialects(::mlir::DialectRegistry &registry) const override {
    
  }

  /// Explicitly declare the TypeID for this class. We declare an explicit private
  /// instantiation because Pass classes should only be visible by the current
  /// library.
  MLIR_DEFINE_EXPLICIT_INTERNAL_INLINE_TYPE_ID(OptimizeGlobalTensorsPassBase<DerivedT>)

protected:
};

//===----------------------------------------------------------------------===//
// RemoveVariablesInSessionInitializerPass
//===----------------------------------------------------------------------===//

template <typename DerivedT>
class RemoveVariablesInSessionInitializerPassBase : public ::mlir::OperationPass<ModuleOp> {
public:
  using Base = RemoveVariablesInSessionInitializerPassBase;

  RemoveVariablesInSessionInitializerPassBase() : ::mlir::OperationPass<ModuleOp>(::mlir::TypeID::get<DerivedT>()) {}
  RemoveVariablesInSessionInitializerPassBase(const RemoveVariablesInSessionInitializerPassBase &other) : ::mlir::OperationPass<ModuleOp>(other) {}

  /// Returns the command-line argument attached to this pass.
  static constexpr ::llvm::StringLiteral getArgumentName() {
    return ::llvm::StringLiteral("tf-saved-model-remove-vars-in-session-initializer");
  }
  ::llvm::StringRef getArgument() const override { return "tf-saved-model-remove-vars-in-session-initializer"; }

  ::llvm::StringRef getDescription() const override { return "Remove variables in tf saved model's session initializer."; }

  /// Returns the derived pass name.
  static constexpr ::llvm::StringLiteral getPassName() {
    return ::llvm::StringLiteral("RemoveVariablesInSessionInitializerPass");
  }
  ::llvm::StringRef getName() const override { return "RemoveVariablesInSessionInitializerPass"; }

  /// Support isa/dyn_cast functionality for the derived pass class.
  static bool classof(const ::mlir::Pass *pass) {
    return pass->getTypeID() == ::mlir::TypeID::get<DerivedT>();
  }

  /// A clone method to create a copy of this pass.
  std::unique_ptr<::mlir::Pass> clonePass() const override {
    return std::make_unique<DerivedT>(*static_cast<const DerivedT *>(this));
  }

  /// Return the dialect that must be loaded in the context before this pass.
  void getDependentDialects(::mlir::DialectRegistry &registry) const override {
    
  }

  /// Explicitly declare the TypeID for this class. We declare an explicit private
  /// instantiation because Pass classes should only be visible by the current
  /// library.
  MLIR_DEFINE_EXPLICIT_INTERNAL_INLINE_TYPE_ID(RemoveVariablesInSessionInitializerPassBase<DerivedT>)

protected:
};
#undef GEN_PASS_CLASSES
#endif // GEN_PASS_CLASSES
#ifdef GEN_PASS_REGISTRATION

//===----------------------------------------------------------------------===//
// DedupBoundInputBindingPass Registration
//===----------------------------------------------------------------------===//

inline void registerDedupBoundInputBindingPassPass() {
  ::mlir::registerPass([]() -> std::unique_ptr<::mlir::Pass> {
    return mlir::tf_saved_model::CreateDedupBoundInputBindingPass();
  });
}

//===----------------------------------------------------------------------===//
// FreezeAssetsPass Registration
//===----------------------------------------------------------------------===//

inline void registerFreezeAssetsPassPass() {
  ::mlir::registerPass([]() -> std::unique_ptr<::mlir::Pass> {
    return mlir::tf_saved_model::CreateFreezeAssetsPass();
  });
}

//===----------------------------------------------------------------------===//
// FreezeGlobalTensorsPass Registration
//===----------------------------------------------------------------------===//

inline void registerFreezeGlobalTensorsPassPass() {
  ::mlir::registerPass([]() -> std::unique_ptr<::mlir::Pass> {
    return ::mlir::tf_saved_model::CreateFreezeGlobalTensorsPass();
  });
}

//===----------------------------------------------------------------------===//
// OptimizeGlobalTensorsPass Registration
//===----------------------------------------------------------------------===//

inline void registerOptimizeGlobalTensorsPassPass() {
  ::mlir::registerPass([]() -> std::unique_ptr<::mlir::Pass> {
    return mlir::tf_saved_model::CreateOptimizeGlobalTensorsPass();
  });
}

//===----------------------------------------------------------------------===//
// RemoveVariablesInSessionInitializerPass Registration
//===----------------------------------------------------------------------===//

inline void registerRemoveVariablesInSessionInitializerPassPass() {
  ::mlir::registerPass([]() -> std::unique_ptr<::mlir::Pass> {
    return mlir::tf_saved_model::CreateRemoveVariablesInSessionInitializerPass();
  });
}

//===----------------------------------------------------------------------===//
// TensorFlowSavedModel Registration
//===----------------------------------------------------------------------===//

inline void registerTensorFlowSavedModelPasses() {
  registerDedupBoundInputBindingPassPass();
  registerFreezeAssetsPassPass();
  registerFreezeGlobalTensorsPassPass();
  registerOptimizeGlobalTensorsPassPass();
  registerRemoveVariablesInSessionInitializerPassPass();
}
#undef GEN_PASS_REGISTRATION
#endif // GEN_PASS_REGISTRATION
